#include <zephyr/types.h>
#include <stddef.h>
#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/logging/log.h>
#include <stdlib.h> 

LOG_MODULE_REGISTER(centralble, LOG_LEVEL_INF);

static void start_scan(void);

static struct bt_conn *default_conn;
static const char *passkey = "minew123";

static const struct bt_conn_le_phy_param phy_pref = {
    .options = BT_CONN_LE_PHY_OPT_NONE,
    .pref_tx_phy = BT_GAP_LE_PHY_1M,
    .pref_rx_phy = BT_GAP_LE_PHY_1M,
};

static const struct bt_le_conn_param *conn_param_1s = BT_LE_CONN_PARAM(800, 800, 0, 400);

struct bt_gatt_exchange_params params = {

};


static void device_found(const bt_addr_le_t *addr, int8_t rssi, uint8_t type,
			 struct net_buf_simple *ad)
{
	char addr_str[BT_ADDR_LE_STR_LEN];
	int err;

	if (default_conn) {
		return;
	}

	/* We're only interested in connectable events */
	if (type != BT_GAP_ADV_TYPE_ADV_IND &&
	    type != BT_GAP_ADV_TYPE_ADV_DIRECT_IND) {
		return;
	}

	bt_addr_le_to_str(addr, addr_str, sizeof(addr_str));

	if (rssi >= -35) {
	printk("Device found: %s (RSSI %d)\n", addr_str, rssi);
	}
	/* connect only to devices in close proximity */
	if (rssi < -40) {
		return;
	}

	if (bt_le_scan_stop()) {
		return;
	}

	err = bt_conn_le_create(addr, BT_CONN_LE_CREATE_CONN,
				conn_param_1s, &default_conn);
	if (err) {
		printk("Create conn to %s failed (%d)\n", addr_str, err);
		start_scan();
	}
	// err = bt_conn_le_phy_update(default_conn, &phy_pref);
    // if (err) {
    //     printk("Failed to set PHY preferences (%d)\n", err);
    // }
	
}

static void start_scan(void)
{
	int err;

	/* This demo doesn't require active scan */
	err = bt_le_scan_start(BT_LE_SCAN_PASSIVE, device_found);
	if (err) {
		printk("Scanning failed to start (err %d)\n", err);
		return;
	}

	printk("Scanning successfully started\n");
}

static void connected(struct bt_conn *conn, uint8_t err)
{
	char addr[BT_ADDR_LE_STR_LEN];


	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (err) {
		printk("Failed to connect to %s (%u)\n", addr, err);

		bt_conn_unref(default_conn);
		default_conn = NULL;

		start_scan();
		return;
	}

	if (conn != default_conn) {
		return;
	}

	printk("Connected: %s\n", addr);

	struct bt_conn_info info;
	err = bt_conn_get_info(conn, &info);
	if (err) {
		printk("Failed to get connection info (%d)\n", err);
		return;
	}
	double connection_interval = info.le.interval * 1.25; // in ms
	uint16_t supervision_timeout = info.le.timeout * 10; // in ms
	LOG_INF("Connection parameters: interval %.2f ms, latency %d intervals, timeout %d ms",
		connection_interval, info.le.latency, supervision_timeout);
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	if (conn != default_conn) {
		return;
	}

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Disconnected: %s (reason 0x%02x)\n", addr, reason);

	bt_conn_unref(default_conn);
	default_conn = NULL;

	start_scan();
}

/* Authentication procedure*/

static void auth_passkey_display(struct bt_conn *conn, unsigned int passkey){
    printk("Passkey for connection %p: %06u\n", conn, passkey);
}

static void auth_passkey_confirm(struct bt_conn *conn, unsigned int passkey) {
    printk("Confirm passkey for connection %p: %06u\n", conn, passkey);
    bt_conn_auth_passkey_confirm(conn);
}

static void auth_cancel(struct bt_conn *conn) {
    printk("Pairing cancelled for connection %p\n", conn);
}

static void auth_pairing_complete(struct bt_conn *conn, bool bonded) {
    printk("Pairing complete for connection %p, bonded: %d\n", conn, bonded);
}

static void auth_pairing_failed(struct bt_conn *conn, enum bt_security_err reason) {
    printk("Pairing failed for connection %p, reason: %d\n", conn, reason);
}

static void auth_passkey_entry(struct bt_conn *conn) {
    int err;

    printk("Enter passkey for connection %p\n", conn);
    char *endptr;
    long passkey_num = strtol(passkey, &endptr, 10);
    if (*endptr != '\0' || passkey_num < 0 || passkey_num > 999999) {
        printk("Invalid passkey\n");
        return;
    }

    err = bt_conn_auth_passkey_entry(conn, passkey_num);
    if (err) {
        printk("Failed to set passkey for connection %p, err: %d\n", conn, err);
    }
}

static void on_security_changed(struct bt_conn *conn, bt_security_t level,
			     enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (!err) {
		LOG_INF("Security changed: %s level %u\n", addr, level);
	} else {
		LOG_INF("Security failed: %s level %u err %d\n", addr, level,
			err);
	}
}
void on_le_param_updated(struct bt_conn *conn, uint16_t interval, uint16_t latency, uint16_t timeout)
{
    double connection_interval = interval*1.25;         // in ms
    uint16_t supervision_timeout = timeout*10;          // in ms
    LOG_INF("Connection parameters updated: interval %.2f ms, latency %d intervals, timeout %d ms", connection_interval, latency, supervision_timeout);
}

void on_le_data_len_updated(struct bt_conn *conn, struct bt_conn_le_data_len_info *info)
{
    uint16_t tx_len     = info->tx_max_len; 
    uint16_t tx_time    = info->tx_max_time;
    uint16_t rx_len     = info->rx_max_len;
    uint16_t rx_time    = info->rx_max_time;
    LOG_INF("Data length updated. Length %d/%d bytes, time %d/%d us", tx_len, rx_len, tx_time, rx_time);
}

static struct bt_conn_auth_cb auth_cb_display = {
    .passkey_display = auth_passkey_display,
	.passkey_entry = auth_passkey_entry,
    .passkey_confirm = auth_passkey_confirm,
    .cancel = auth_cancel,
    //.pairing_confirm = auth_pairing_complete,
};


BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
	.security_changed = on_security_changed,
	.le_param_updated  = on_le_param_updated,
	.le_data_len_updated  = on_le_data_len_updated,

};

int main(void)
{
	int err;

	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return 0;
	}

	printk("Bluetooth initialized\n");
	bt_conn_auth_cb_register(&auth_cb_display);

	start_scan();
	return 0;
}
